#!/bin/bash

. /etc/os-release

rpm_dest=""
docker_image_dest=""

include_docker=1
include_kube=1

PIGZ_RPM="pigz-2.3.3-1.el7.centos.x86_64.rpm"

if [[ "$ID" == rhel && ! -e "$PIGZ_RPM" ]]; then
	echo "$PIGZ_RPM must be in the current working directory to properly"
	echo "gather the files required for an offline installation on this"
	echo "system."
fi

usage() {
#===============================================================================
	cat << EOF >&2
Usage: $0 [-nd] [-nk] -dest [parent directory]

-dest          Store RPMs in <parent directory>/rpms, docker image files in
               <parent directory>/docker-images, and use
               <parent directory>/installroot as the installroot temporary
               directory.

-nd            The system where the solution will be installed already has
               docker installed, no need to pull the docker rpms.

-nk            The system where the solution will be installed already has
               kubernetes installed, no need to pull the kubernetes rpms.

Example $0 -dest /fcimedia/offline-files

/fcimedia/offline-files/rpms - RPMs required for offline installation.
/fcimedia/offline-files/docker-images - docker image files required for offline
                                        installation.
EOF
#===============================================================================
}

process_args() {
	local error_count=0
	while [[ $# -gt 0 ]]; do
		local key="$1"; shift
		case $key in
		-nd)
			include_docker=0
			;;
		-nk)
			include_kube=0
			;;
		-dest)
			rpm_dest="$1/rpms"
			docker_image_dest="$1/docker-images"
			shift
			;;
		*)
			echo "Unrecognized parameter: $key"
			error_count=$((error_count + 1))
			shift
			;;
		esac
	done

	if [[ $rpm_dest == "" ]]; then
		echo "Destination location for downloaded files required. Set with:"
		echo "  o -dest [ parent directory ]"
		error_count=$((error_count + 1))
	fi

	return $error_count
}

is_installed() {
	local -r package="$1"
	local rc=0

	echo -n "Checking for $package..."
	rpm -qi "$package" > /dev/null 2>&1
	rc=$?

	if [[ $rc -eq 0 ]]; then
		echo "  Already installed"
	else
		echo "  Need to install"
	fi

	return $rc
}

install_dependencies() {
	local required_packages=( curl docker yum-utils )
	local packages=""
	local missing=0

	for package in "${required_packages[@]}"; do
		is_installed "$package" || {
			packages="$packages $package"
			missing=$((missing + 1))
		}
	done

	if [[ $missing -gt 0 ]]; then
		yum -y install $packages || { echo "yum install of required packages failed."; exit 1; }
	fi
}

download_rpms() {
	local urls=()
	local kube_rpms=( \
		"https://packages.cloud.google.com/yum/pool/35ea034b2efccf2529cca8ed44f1bdcc0c3b26f0139694d8cbea315077a1bf6e-kubeadm-1.10.1-0.x86_64.rpm"
		"https://packages.cloud.google.com/yum/pool/b8f0cc3bc85e8614f0340547b14ca5377410afd087f51077410945a594f0b71b-kubectl-1.10.1-0.x86_64.rpm"
		"https://packages.cloud.google.com/yum/pool/4cad7573e2617d903a12f5d318f597ec47bc0e67792a04f67ffee3e06c0ad373-kubelet-1.10.1-0..x86_64.rpm"
		"https://packages.cloud.google.com/yum/pool/fe33057ffe95bfae65e2f269e1b05e99308853176e24a4d027bc082b471a07c0-kubernetes-cni-0.6.0-0.x86_64.rpm"
	)
	local docker_rpms=( \
		"https://download.docker.com/linux/centos/7/x86_64/stable/Packages/docker-ce-18.03.1.ce-1.el7.centos.x86_64.rpm"
	)
	local other_files=( \
		"https://github.com/projectcalico/calicoctl/releases/download/v1.6.1/calicoctl"
		"https://storage.googleapis.com/kubernetes-helm/helm-v2.9.0-linux-amd64.tar.gz"
	)
	local pigz=""
	if [[ $include_docker -eq 1 ]]; then
		urls+=("${docker_rpms[@]}")
	fi
	if [[ $include_kube -eq 1 ]]; then
		urls+=("${kube_rpms[@]}")
	fi
	urls+=("${other_files[@]}")

	if [[ "$ID" == rhel ]]; then
		pigz="$PIGZ_RPM"
	fi

	mkdir -p "$rpm_dest" || { echo "Failed to create required directories."; exit 1; }

	local curl_err=0
	local output_redir=""
	local i=0 url_count=${#urls[@]}
	for (( i=0; i<${url_count}; i++ )); do
		output_redir="$output_redir -O"
	done
	pushd "${rpm_dest}"
	echo "Downloading $url_count files."
	curl -L $output_redir "${urls[@]}" || curl_err=1
	popd
	if [[ "$curl_err" -ne 0 ]]; then
		echo "Failed to download the required RPMs for an offline solution."
		exit 1
	fi
	
	if [[ "$pigz" != "" && ! -e ${rpm_dest}/pigz* ]]; then
		cp "${pigz}" "${rpm_dest}"
	fi
}

save_docker_images() {
	mkdir -p "$docker_image_dest"
	local installed_docker=0
	local started_docker=0

	systemctl status docker > /dev/null 2>&1
	if [[ $? -ne 0 ]]; then
		systemctl start docker || { echo "Cannot start docker."; exit 1; }
		started_docker=1
	fi

	local -a failed_images=()
	while read image; do
		docker pull "$image" || { echo "Failed to pull $image"; failed_images+=("$image"); break; }
		docker save "$image" > "$docker_image_dest/$(echo "${image}.tar" | tr '/:' '--')" || \
			{ echo "Failed to save $image"; failed_images+=("$image"); break; }
	done << EOF
registry:2
centos:7
gcr.io/google_containers/defaultbackend:1.4
gcr.io/google_containers/pause-amd64:3.0
gcr.io/kubernetes-helm/tiller:v${helm_version}
k8s.gcr.io/etcd-amd64:3.1.12
k8s.gcr.io/k8s-dns-dnsmasq-nanny-amd64:1.14.8
k8s.gcr.io/k8s-dns-kube-dns-amd64:1.14.8
k8s.gcr.io/k8s-dns-sidecar-amd64:1.14.8
k8s.gcr.io/kube-apiserver-amd64:v${kube_version}
k8s.gcr.io/kube-controller-manager-amd64:v${kube_version}
k8s.gcr.io/kube-proxy-amd64:v${kube_version}
k8s.gcr.io/kube-scheduler-amd64:v${kube_version}
k8s.gcr.io/pause-amd64:3.0
quay.io/calico/cni:v2.0.0
quay.io/calico/kube-controllers:v2.0.0
quay.io/calico/node:v3.0.2
quay.io/coreos/etcd:v3.1.10
quay.io/kubernetes-ingress-controller/nginx-ingress-controller:0.12.0
EOF

	if [[ $started_docker -eq 1 ]]; then
		systemctl stop docker
	fi

	if [[ "${#failed_images[@]}" -gt 0 ]]; then
		echo "Failed to download the following docker images:"
		echo "${failed_images[@]}"
		exit 1
	fi
}

main() {
	if ! process_args "$@"; then
		usage
		exit 1
	fi
	install_dependencies

	download_rpms
	save_docker_images
}

main "$@"

