#!/usr/bin/env bash

. gettext.sh

proxy_chart="nginx-ingress-controller-1.0.3-APAR-PH01524.tgz"
PIGZ_RPM="pigz-2.3.3-1.el7.centos.x86_64.rpm"

# Set variables
SSHKEY_NAME="kuberentes_manager_rsa"

shopt -s expand_aliases

# Check that script is run as root user
if [ "$EUID" -ne 0 ]; then
	log.error "$(gettext -s "error_require_root")"
	exit
fi

# If LC_CTYPE is set, it stops the translations from working.
if [[ $LC_CTYPE != "" ]]; then
	unset LC_CTYPE
fi

# Capture the current working directory
SCRIPT_DIR=$(cd -P -- "$(dirname -- "$0")" && pwd -P)
SCRIPT_NAME=$(basename "$0")
export TEXTDOMAINDIR="$SCRIPT_DIR/locale"
export TEXTDOMAIN="install.sh"

# Define variables to create a lock on the script
LOCKFILE="${SCRIPT_DIR}/$SCRIPT_NAME.lock"
LOCKFD=99

# Lock functions
_lock()			{ echo $$>"$LOCKFILE"; flock -"$1" "$LOCKFD"; }
_remove_lock()		{ _lock u; _lock xn && rm -f "$LOCKFILE"; }
_init_lock()		{ eval "exec $LOCKFD>\"$LOCKFILE\""; trap _remove_lock EXIT; }
_exclusivelock()	{ _lock xn; }
_alreadylocked()	{ printf "\nUnable to attain an exclusive lock on %s.  Exiting.\nPlease verify that this script isn't already running.\n\n" "$SCRIPT_DIR/${0##*/}"; exit 1; }

# Create lock
_init_lock
_exclusivelock || _alreadylocked

# Define properties files
PROPERTY_FILE="install.properties"

# HOST_TYPE IP_ADDRESS HOSTNAME_FQDN HOSTNAME_SHORT
HOSTNAME_FILE="install.hosts.properties"

configure_nodes=0

# Define functions

# shellcheck disable=SC2034
declare -r logging_info=$(gettext -n -s "log_info")
# shellcheck disable=SC2034
declare -r logging_warning=$(gettext -n -s "log_warning")
# shellcheck disable=SC2034
declare -r logging_error=$(gettext -n -s "log_error")

alias log.info='logging $logging_info $LINENO $INSTALL_STEP; echo '
alias log.warn='logging $logging_warning $LINENO $INSTALL_STEP; echo '
alias log.error='logging $logging_error $LINENO $INSTALL_STEP; echo '
alias log.header='log-header $LINENO $INSTALL_STEP '

logging() {
	# shellcheck disable=SC2034
	local -r log_type="${1}"; shift
	# shellcheck disable=SC2034
	local -r line_no="${1}"; shift
	# shellcheck disable=SC2034
	local -r install_step="${1}"; shift
	# shellcheck disable=SC2034
	local -r date="$(date --rfc-3339=s)"

	eval_gettext "log_record_4 \$date \$log_type \$line_no \$install_step"
}

log-header() {
	local -r line_no="${1}"; shift
	local -r install_step="${1}"; shift
	local -r key="${1}"; shift
	local -r lines="$(gettext -s log_header_lines)"

	logging "$logging_info" "$line_no" "$install_step"
	echo "$lines"

	logging "$logging_info" "$line_no" "$install_step"
	eval_gettext "$key"
	echo

	logging "$logging_info" "$line_no" "$install_step"
	echo "$lines"
}

usage() {
	local name=$0
	echo
	eval_gettext "usage_1 \$name"
	echo
	exit 1
}

convert-seconds() {
	local hours minutes seconds

	(( hours   = ${1}/3600 ))
	(( minutes = (${1}%3600)/60 ))
	(( seconds = ${1}%60 ))

	eval_gettext "elapsed_time_3 \$hours \$minutes \$seconds"
}

element-in() {
	local e match="$1"
	shift
	for e; do [[ $e == "$match" ]] && return 0; done
	return 1
}

element-in-array() {
	local match="$1"
	local -a elements=("${!2}")

	for e in "${elements[@]}"; do
		# For some reason, can't quote either of these or
		# worker.* != worker.1
		# shellcheck disable=SC2053
		[[ $match == $e ]] && return 0
	done
	return 1
}

text-in-string() {
	[ -z "${2##*$1*}" ];
}

read-host-roles() {
	grep "\.fqdn" "$HOSTNAME_FILE" | grep -v "^#" | cut -d= -f 1 | sed -e 's/.fqdn//g'
}

read-host-fqdn() {
	grep "\.fqdn" "$HOSTNAME_FILE" | grep -v "^#" | cut -d= -f 2
}

# input is an array of host types in order of preference.
# output fqdn of the host
find-best-host() {
	local -a target_host_types=("${!1}")
	local rc=0
	local fqdn=""

	while [[ "$fqdn" == "" && "${#target_host_types[@]}" -gt 0 ]]; do
		local target_role="${target_host_types[0]}"
		fqdn="$(get-property-value "${HOSTNAME_FILE}" "${target_role}.fqdn")"
		target_host_types=("${target_host_types[@]:1}")
	done

	echo "$fqdn"

	return $rc
}

get-property-value() {
	local -r propertyFile="$1"; shift
	local property="$1"; shift
	local -r default_value="$1"; shift
	local value=""
	local rc=0 

	while true
	do
		[[ "$propertyFile" == "" || "$property" == "" ]] && rc=1 && break 

		property="${property%%[[:space:]]}"		
		property="${property##[[:space:]]}"	

		[[ ${property} == "" ]] && rc=2 && break
		[[ ! -f "${propertyFile}" ]] && rc=3 && break

		value=$(sed '/^\#/d' "${propertyFile}" | grep "^[[:space:]]*${property}[=:[:space:]]" | tail -n 1 )		
		[[ ${value} == ""  && $default_value != "" ]] && value="${property} = ${default_value}"
		[[ ${value} == "" ]] && rc=4 && break

		# Remove leading/trailing white space 
		value="${value%%[[:space:]]}"		
		value="${value##[[:space:]]}"	
		# Isolate value (split on ':' or '=')
		value=$(echo "${value}" | sed 's/[^:=]*//')
		# delete everything after the first # (comment)
		value=${value%%"#"*}
		# Remove leading value whitespace
		value=$(echo "${value}" | sed 's/^[:=][[:space:]]*//')
		# Remove "Windows" CR character (if any) from value.
		value=$(echo "${value}" | sed 's|\r$||')
		break
	done
	echo -n "${value}"
	return $rc
}

find-license-dir() {
	local target="license/Softcopy"
	local dir="$SCRIPT_DIR"

	while [[ -e "$dir" ]]; do
		if [[ -e "$dir/$target" ]]; then
			echo -n "$dir/$target"
			return 0
		fi
		dir="${dir%/*}"
	done
	return 1
}

find-license-file-from-var() {
	local -r license_dir="$1"; shift
	local -r prefix="$1"; shift
	local suffix="$1"; shift
	local file="$license_dir/$prefix$suffix"

	# lang_region.encoding
	[[ -e "$file" ]] && { echo -n "$file"; return 0; }

	# lang_region
	suffix=${suffix%\.*}
	file="$license_dir/$prefix$suffix"
	[[ -e "$file" ]] && { echo -n "$file"; return 0; }

	# lang
	suffix=${suffix%_*}
	file="$license_dir/$prefix$suffix"
	[[ -e "$file" ]] && { echo -n "$file"; return 0; }

	return 1
}

find-license-file() {
	local -r prefix="$1"; shift
	local -r license_dir=$(find-license-dir)
	local -r -i license_dir_ok=$?

	[[ $license_dir_ok -eq 0 ]] || { log.error "$(gettext -s "license_dir_not_found")" && echo; return 1; } 

	local license_file=$(find-license-file-from-var "$license_dir" "$prefix" "$LC_ALL")
	[[ $license_file != "" ]] && { echo -n "$license_file"; return 0; }

	license_file=$(find-license-file-from-var "$license_dir" "$prefix" "$LANG")
	[[ $license_file != "" ]] && { echo -n "$license_file"; return 0; }

	license_file=$(find-license-file-from-var "$license_dir" "$prefix" "en_US")
	[[ $license_file != "" ]] && { echo -n "$license_file"; return 0; }

	return 1
}

show-license-acceptance-instructions() {
	local -r license_key="${1}"; shift

	log.header "$license_key"
	gettext -s "license_acceptance_instructions"
	read _
	echo
	return 0
}

prompt-for-license-acceptance() {
	local -r license_key="${1}"; shift
	local -r license_file="${1}"; shift
	local vi_cmd=$(command -v vi)
	local -r yes=$(gettext -n -s "license_acceptance_yes")
	local -r no=$(gettext -n -s "license_acceptance_no")
	local rc=1

	[[ $vi_cmd != "" ]] || { log.error "$(gettext -s "cannot_display_license_file_vi_not_found")"; return 1; }

	show-license-acceptance-instructions "$license_key"
	"$vi_cmd" -R "$license_file"
	echo; echo
	eval_gettext "prompt_for_license_acceptance_2 \$yes \$no"
	while true; do
		read response
		[[ ${response} == "$yes" ]] && { rc=0; break; }
		[[ ${response} == "$no" ]] && { break; }
		echo
		eval_gettext "invalid_input_license_acceptance_3 \$response \$yes \$no"
	done

	return $rc
}

is_install_step_done() {
	local -r install_step="${1}"; shift
	local -r lock_file="${1}"; shift
	local rc=0

	if [[ -f "$lock_file" ]]; then
		# shellcheck disable=SC2034
		local -r note="$(cat "$lock_file")"
		log.info "$(eval_gettext "log_step_already_completed_2 \$install_step \$note")"
		rc=1
	fi
	return $rc
}

install_step_done() {
	local -r lock_file="${1}"; shift
	# shellcheck disable=SC2034
	local -r when=$(date)

	eval_gettext "log_step_completion_note_1 \$when" > "$lock_file"
}

acceptLicenseAgreement() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local -r license_file=$(find-license-file LA_)
	local -r -i license_file_rc=$?

	[[ $license_file_rc -eq 0 ]] || { echo "$license_file"; return 1; }
	prompt-for-license-acceptance "license_agreement" "$license_file"
	[[ $? -eq 0 ]] || error_count=$((error_count + 1))

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

acceptLicenseInformation() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local -r license_file=$(find-license-file LI_)
	local -r -i license_file_rc=$?

	[[ $license_file_rc -eq 0 ]] || { echo "$license_file"; return 1; }
	prompt-for-license-acceptance "license_information" "$license_file"
	[[ $? -eq 0 ]] || error_count=$((error_count + 1))

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

installExpect() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	# Install expect
	log.header "step_install_expect"
	if  rpm -q expect; then
		log.info "$(gettext -s "step_install_expect_already_done")"
		installexpectresult=0
	else
		log.info "$(gettext -s "step_install_expect_installing")"
		yum -y install expect
		installexpectresult=$?
	fi

	[[ $installexpectresult -eq 0 ]] && install_step_done "$lockfile"
	return $installexpectresult
}

declare HOST_PASSWD=""
setupSSHTrust () {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	local my_hostname=$(hostname -f)
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0
	
	# Generate SSH public/private key pair
	log.header "step_ssh_trust"

	log.info "$(gettext -s "step_ssh_trust_remove_old_key")"
	# Remove existing id_rsa if it exists
	if [[ -f /root/.ssh/id_rsa ]]; then 
		/usr/bin/rm -f /root/.ssh/id_rsa && /usr/bin/rm -f /root/.ssh/id_rsa.pub 
	fi

	# Generate a new id_rsa key
	log.info "$(gettext -s "step_ssh_trust_generate_new_key")"
	ssh-keygen -q -t rsa -N "" -f /root/.ssh/id_rsa || { log.error "$(gettext -s "step_ssh_trust_generate_new_key_failed")" && echo; return 1; }

	# Setup trust from kubernetes manager to all other nodes
	log.info "$(gettext -s "step_ssh_trust_uploading_key")"
	local HOST_TYPE
	while read HOST_TYPE; do
		local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
		HOST_PASSWD="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.root_password")"
		# Only execute on remote hosts
		if [[ $HOSTNAME_FQDN != $my_hostname ]]; then
			log.info "$(eval_gettext "step_ssh_trust_create_trust_3 \$my_hostname \$HOST_TYPE \$HOSTNAME_FQDN")"
	
			# Extract properties from property file (required for runExpect step)

			# Upload public key and update authorized_keys 	
			runExpect 'ssh -q -n' "root@$HOSTNAME_FQDN" "mkdir -p /root/.ssh/" || { log.error "$(eval_gettext "step_ssh_trust_remote_create_dir_failed \$HOSTNAME_FQDN")" && echo; error_count=$((error_count + 1)); }
			runExpect 'ssh -q -n' "root@$HOSTNAME_FQDN" "rm -f /root/.ssh/authorized_keys" || { log.error "$(eval_gettext "step_ssh_trust_remote_rm_authorized_keys_failed \$HOSTNAME_FQDN")" && echo; error_count=$((error_count + 1)); } 
			runExpect scp /root/.ssh/id_rsa.pub "root@$HOSTNAME_FQDN:/root/.ssh/${SSHKEY_NAME}.pub" || { log.error "$(eval_gettext "step_ssh_trust_remote_scp_pub_key_failed \$SSHKEY_NAME \$HOSTNAME_FQDN")" && echo; error_count=$((error_count + 1)); } 
			runExpect ssh-copy-id '-i /root/.ssh/id_rsa.pub' "root@$HOSTNAME_FQDN" || { log.error "$(eval_gettext "step_ssh_trust_remote_ssh_copi_id_failed \$HOSTNAME_FQDN")" && echo; error_count=$((error_count + 1)); } 
		else
			log.info "$(eval_gettext "step_ssh_trust_local \$my_hostname \$HOST_TYPE \$HOSTNAME_FQDN")"

			# Upload public key and update authorized_keys 	
			runExpect ssh-copy-id '-i /root/.ssh/id_rsa.pub' "root@$HOSTNAME_FQDN" || { log.error "$(eval_gettext "step_ssh_trust_remote_ssh_copi_id_failed \$HOSTNAME_FQDN")" && echo; error_count=$((error_count + 1)); } 
		fi
	done < <(read-host-roles)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

sshTest() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	# Checks if ssh root trusts are enabled (servers listed in install.hosts.properties)
	log.header "step_ssh_test \$HOSTNAME_FILE"
	while read HOSTNAME_FQDN; do
		sshtest=$(ssh -q -n -o StrictHostKeyChecking=no -o BatchMode=yes -o ConnectTimeout=5 "root@$HOSTNAME_FQDN" echo ok 2>&1)
		if [[ $sshtest == "ok" ]]; then		
			log.info "$(eval_gettext "step_ssh_test_pass \$HOSTNAME_FQDN")"
		else 
			log.warn "$(eval_gettext "step_ssh_test_fail \$HOSTNAME_FQDN")"
			error_count=$((error_count + 1))
		fi
	done < <(read-host-fqdn)

	if [[ $error_count -gt 0 ]]; then
		log.error "$(gettext -s "step_ssh_test_fail_summary")"
		log.error "$(gettext -s "step_ssh_test_fail_instructions")"
		log.error "$(eval_ngettext "step_ssh_test_fail \$error_count" "step_ssh_test_fails \$error_count" $error_count)"
	else 
		log.info "$(gettext -s "step_ssh_test_success")"
		install_step_done "$lockfile"
	fi
	return $error_count
}

runExpect() {
	# Uses expect for remote logins before root ssh trust is established
	RUN_EXPECT=$(expect -c "
		log_user 0
		spawn -noecho $1 $2 $3
			expect {
					-re \"$(gettext -s "run_expect_are_you_sure")\" { send \"$(gettext -s "run_expect_yes")\n\"; exp_continue }
						\"$(gettext -s "run_expect_password")\" {send -- \"$HOST_PASSWD\r\";exp_continue}
						eof
					} 
		log_user 1")

	rc=$?

	if [[ "$RUN_EXPECT" != "" ]]; then
		echo "$RUN_EXPECT"
	fi
	return $rc
}

check-property-file() {
	# shellcheck disable=SC2034
	local -r directory="${1}"; shift
	local -r filename="${1}"; shift
	local rc=0

	if [[ ! -f $filename ]]; then
		log.error "$(eval_gettext "property_file_not_found \$directory \$filename")"
		rc=1
	else
		log.info "$(eval_gettext "property_file_found \$directory \$filename")"
	fi

	return $rc
}


get_directory_partition() {
	local directory="$1"; shift
	local rc=1
	local fstype

	if [[ $directory != /* ]]; then
		echo "must be absolute path"
		return 1
	fi

	while [ $rc -ne 0 ]; do
		fstype=$(df --output=fstype "$directory" 2>/dev/null | tail -1)
		if [[ "$fstype" != "" ]]; then
			rc=0
			echo "$directory"
		else
			directory="${directory%/*}"
		fi
	done
	return $rc
}

get_directory_fstype() {
	local directory="$1"; shift
	local rc=1
	local fstype

	if [[ $directory != /* ]]; then
		echo "must be absolute path"
		return 1
	fi

	while [ $rc -ne 0 ]; do
		fstype=$(df --output=fstype "$directory" 2>/dev/null | tail -1)
		if [[ "$fstype" != "" ]]; then
			rc=0
			echo "$fstype"
		else
			directory="${directory%/*}"
		fi
	done
	return $rc
}

is_xfs_settings_ok() {
	local directory="$1"; shift
	local rc=1
	local -r partition=$(get_directory_partition "$directory")

	if [[ $? -eq 0 ]]; then
		local -r fstype=$(get_directory_fstype "$partition")
		if [[ $? -eq 0 ]]; then
			if [[ "$fstype" == "xfs" ]]; then
				local xfs_info=$(xfs_info "$partition" | tr '\n' ' ')
				xfs_info="${xfs_info##* ftype=}"
				xfs_info="${xfs_info%% *}"
				if [[ "$xfs_info" == "1" ]]; then
					rc=0
				fi
			else
				rc=0
			fi
		fi
	fi

	return $rc
}

precheckFiles() {	
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	# Check for required properties files
	log.header "step_precheck"

	if ! check-property-file "$SCRIPT_DIR" "$HOSTNAME_FILE"; then
		error_count=$((error_count + 1))
	else
		manager="$(get-property-value "${HOSTNAME_FILE}" "manager.fqdn")"
		if [[ $manager == "" ]]; then
			log.error "$(eval_gettext "step_precheck_require_manager \$HOSTNAME_FILE")"
			error_count=$((error_count + 1))
		elif [[ $manager != "$(hostname -f)" ]]; then
			log.error "$(eval_gettext "step_precheck_require_run_on_manager \$manager")"
			error_count=$((error_count + 1))
		fi

	while read HOSTNAME_FQDN; do
		if [[ ! $HOSTNAME_FQDN =~ ^[a-z0-9]+[-a-z0-9]+[\.[a-z0-9]+[-a-z0-9]+]*$ ]]; then
			log.error "$(eval_gettext "step_precheck_calico_invalid_hostname $HOSTNAME_FQDN")"
			error_count=$((error_count + 1))
		fi
	done < <(read-host-fqdn)

	fi

	if ! is_xfs_settings_ok /var/lib/docker; then
		log.error "$(gettext -s "step_precheck_bad_xfs_ftype")"
		error_count=$((error_count + 1))
	fi

	if ! check-property-file "$SCRIPT_DIR" "$PROPERTY_FILE"; then
		error_count=$((error_count + 1))
	else
		local -r docker_type=$(get-property-value "${PROPERTY_FILE}" "docker.type" "ce")
		if [[ $docker_type == "ee" ]]; then
			local -r ee_url=$(get-property-value "${PROPERTY_FILE}" "docker.ee.url")
			if [[ $ee_url == "" ]]; then
				log.error "$(gettext -s "step_precheck_docker_ee_url_missing")"
				error_count=$((error_count + 1))
			fi
		elif [[ $docker_type == "ce" ]]; then
			:
		else
			log.error "$(eval_gettext "step_precheck_invalid_docker_type \$docker_type \$PROPERTY_FILE")"
			error_count=$((error_count + 1))
		fi

		local kube_version="$(get-property-value "${PROPERTY_FILE}" "version.kubernetes" "1.10.1")"
		local kube_cni_version="$(get-property-value "${PROPERTY_FILE}" "version.kubernetes.cni" "0.6.0")"
		local helm_version="$(get-property-value "${PROPERTY_FILE}" "version.helm" "2.9.0")"

		local -r external_docker_reg=$(get-property-value "${PROPERTY_FILE}" "external.docker.registry.url")
		local external_nfs_server=$(get-property-value "${PROPERTY_FILE}" "external.nfsserver")

		if rpm -q kubelet; then
			# kubernetes already installed, make sure a bunch of other things
			# are specified.
			if [[ $external_docker_reg == "" ]]; then
				log.error "$(gettext "step_precheck_kube_installed_need_docker_reg")"
				error_count=$((error_count + 1))
			fi
			if [[ $external_nfs_server == "" ]]; then
				log.error "$(gettext "step_precheck_kube_installed_need_nfs_server")"
				error_count=$((error_count + 1))
			fi
		fi

		if [[ $external_docker_reg != "" ]]; then
			local -r reg_is_secure=$(get-property-value "${PROPERTY_FILE}" "external.docker.registry.isSecure")
			if [[ $reg_is_secure == "" ]]; then
				log.error "$(gettext -s "step_precheck_external_docker_reg_security_indicator_missing")"
				error_count=$((error_count + 1))
			fi
		fi

		bash "${SCRIPT_DIR}/prepare/verifyPrereqPackages.txt"
		if [[ $? -ne 0 ]]; then
			error_count=$((error_count + 1))
		fi

		local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")
		if [ ! -d "$media_dir" ]; then
			log.error "$(eval_gettext "step_precheck_bad_media_dir \$media_dir")"
			error_count=$((error_count + 1))
		else
			pushd "$media_dir"
			log.info "$(gettext -s "step_precheck_validating_media")"
			for sha1_file in *.sha1; do
				dos2unix "$sha1_file"
				sha1sum -c "$sha1_file"
				if [[ $? != 0 ]]; then
					log.error "$(eval_gettext "step_precheck_validating_media_failed \$sha1_file")"
					error_count=$((error_count + 1))
				fi
			done
			popd
		fi
	fi

	if [[ "$error_count" != "0" ]]; then
		log.error "$(eval_ngettext "step_precheck_error_count \$error_count" "step_precheck_error_counts \$error_count" $error_count)"
	fi

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

extractArchives() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_extract_archives"
	local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")
	mkdir -p "$media_dir/archives"
	
	for tar_file in "$media_dir"/*.tar; do
		log.info "$(eval_gettext "step_extract_archives \$tar_file")"
		tar xf "$tar_file" -C "$media_dir/archives"
		if [[ $? != "0" ]]; then
			log.error "$(eval_gettext "step_extract_archives_failed \$tar_file")"
			error_count=$((error_count + 1))
		fi
	done

	# The repo_tar might be split if it's too big...
	local -r repo_tar=$(get-property-value "${PROPERTY_FILE}" "docker.registry.file")
	if [[	! -e "$media_dir/archives/$repo_tar" && \
		-e "$media_dir/archives/$repo_tar.partaa" ]]; then
		cat "$media_dir/archives/$repo_tar.part"* > "$media_dir/archives/$repo_tar" && \
		/bin/rm "$media_dir/archives/$repo_tar".part*
	fi

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

validateHostTypes() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	# Validate host types in HOSTNAME_FILE
	log.header "step_validate_host_types"
	local -r -a valid_host_types=( manager nfs docker worker* )
	while read HOST_TYPE; do
	# Check that all host types are valid
		element-in-array "$HOST_TYPE" valid_host_types[@]
		if [[ $? -ne 0 ]]; then
    		log.error "$(eval_gettext "step_validate_host_types_invalid \$HOST_TYPE \$HOSTNAME_FQDN") ${valid_host_types[*]}"
 			error_count=$((error_count + 1))
 		else
    		log.info "$(eval_gettext "step_validate_host_types_valid \$HOST_TYPE \$HOSTNAME_FQDN")"
 		fi
	done < <(read-host-roles)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

ping-target() {
	local ping_target="${1}"; shift
	local rc=0

	ping -q -c4 "$ping_target" > /dev/null
	rc=$?

	if [[ $rc -eq 0 ]]; then
		log.info "$(eval_gettext "step_ping_test_pass \$ping_target")"
	else
		log.error "$(eval_gettext "step_ping_test_fail \$ping_target")"
	fi

	return $rc
}

pingTest() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	# Simple ping test (servers listed in $HOSTNAME_FILE)
	log.header "step_ping_test"
	local hostname=$(hostname -f)
	local ips_for_hostname
	while read HOST_TYPE; do
		local IP_ADDRESS="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.ip")"
		local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
		ping-target "$IP_ADDRESS" || error_count=$((error_count + 1))
		ping-target "$HOSTNAME_FQDN" || error_count=$((error_count + 1))
		if [[ "$HOST_TYPE" == "manager" ]]; then
			local EXTERNAL_IP_ADDRESS="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.external_ip")"
			if [[ $EXTERNAL_IP_ADDRESS != "" ]]; then
				ping-target "$EXTERNAL_IP_ADDRESS" || \
					error_count=$((error_count + 1))
			fi
		fi

		if [[ "$HOSTNAME_FQDN" != "$hostname" ]]; then
			ips_for_hostname=$(getent ahostsv4 "$HOSTNAME_FQDN" | awk '{print $1}' | sort -u)
		else
			ips_for_hostname=$(ip -4 address | grep global | awk '{print $2}' | awk -F/ '{print $1}' | sort -u)
		fi
		if [ ! -z "${ips_for_hostname##*$IP_ADDRESS*}" ]; then
			log.error "$(eval_gettext "step_ping_test_hostname_ip_mismatch \$HOSTNAME_FQDN \$IP_ADDRESS")"
			log.error "$(eval_ngettext "step_ping_test_hostname_ip_matches \$ips_for_hostname")"
			error_count=$((error_count + 1))
		fi
	done < <(read-host-roles)
	
	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

checkUMASK () {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	# Check umask setting on all servers
	log.header "step_validate_umask"
	local hostname=$(hostname -f)
	while HOSTNAME_FQDN; do
		# Check on remote hosts
		if [[ "$HOSTNAME_FQDN" != "$hostname" ]]; then
			CHECK_UMASK=$(ssh -n "root@$HOSTNAME_FQDN" 'umask')
		else
			# Check local umask setting
			CHECK_UMASK="$(umask)"
		fi
	
		# Check that umask is set to 0022
		if [[ $CHECK_UMASK != "0022" ]]; then
			log.error "$(eval_gettext "step_validate_umask_invalid \$CHECK_UMASK \$HOSTNAME_FQDN")"
			error_count=$((error_count + 1))
		else
			log.info "$(eval_gettext "step_validate_umask_valid \$HOSTNAME_FQDN")"
		fi
	done < <(read-host-fqdn)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

process-prereq-list-local() {
	# Create prereq script for remote execution
	prereqScriptPath="${SCRIPT_DIR}/prepare/_generatedScripts"
	prereqScript="install_prereqs_$HOST_TYPE.sh"
	generate-prereq-script || { log.error "$(gettext -s "generate_script_failed")" && echo; return 1; }
	
	# Run $PrereqScript
	chmod u+x "$prereqScriptPath/$prereqScript"
	"$prereqScriptPath/$prereqScript"
	return 0
}

process-prereq-list-remote() {
	# Create prereq script for remote execution
	prereqScriptPath="${SCRIPT_DIR}/prepare/_generatedScripts"
	prereqScript="install_prereqs_$HOST_TYPE.sh"
	generate-prereq-script || { log.error "$(gettext -s "generate_script_failed")" && echo; return 1; }

	# Upload files to host and run
	scp "$prereqScriptPath/$prereqScript" "root@$HOSTNAME_FQDN:~"
	# shellcheck disable=SC2029
	ssh -n "root@$HOSTNAME_FQDN" "chmod u+x $prereqScript && ./$prereqScript; rm $prereqScript"
	return 0
}

generate-prereq-script() {
	local rc=0 
	local dest="$prereqScriptPath/$prereqScript"

	# Initialize bash script
	echo "#!/bin/sh" > "$dest"
	
	# Update the script with pre-req steps
	local -i list_iter=0
	while [[ $list_iter -lt $list_size ]]; do
		local file="${PREREQ_LIST[$list_iter]}"
		cat <<EOF >> "$dest"
#############################################
# $file
#############################################
EOF
		cat "$file" >> "$dest"
		rc=$?
		if [[ $rc -ne 0 ]]; then
			log.error "$(eval_gettext "generate_script_failed_detail \$file \$dest")"
			break 2
		fi
		list_iter+=1
		echo "" >> "$dest"
	done
	return $rc
}

updateOSconfig() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_update_os"
	# Deploy pre-reqs on servers in HOSTNAME_FILE
	# Create required directory
	local script_gen_dir="${SCRIPT_DIR}/prepare/_generatedScripts"
	mkdir -p "${script_gen_dir}" || { log.error "$(eval_gettext "step_update_os_create_dir_failed \$script_gen_dir")" && echo; return 1; }
	local hostname=$(hostname -f)
	while read HOST_TYPE; do
		local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
		list_size=0
		unset PREREQ_LIST
		log.info "$(eval_gettext "setup_update_os_server \$HOSTNAME_FQDN \$HOST_TYPE")"
		if [[ $HOST_TYPE == manager ]] || [[ $HOST_TYPE == worker* ]]; then	
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/configure_kernel_parameters.txt"
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/disable_swap.txt"	
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/enable_ntpd.txt"			
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/disable_selinux.txt"
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/disable_fw.txt"
		elif [[ $HOST_TYPE == docker ]] || [[ $HOST_TYPE == nfs ]]; then
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/enable_ntpd.txt"			
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/disable_selinux.txt"
			PREREQ_LIST[list_size++]="${SCRIPT_DIR}/prepare/disable_fw.txt"
		fi
		if [[ $list_size -gt 0 ]]; then
			# Check if installing locally
			if [[ $HOSTNAME_FQDN == "$hostname" ]]; then
				process-prereq-list-local || { log.error "$(eval_gettext "setup_update_os_failed \$HOSTNAME_FQDN \$HOST_TYPE")" && echo; return 1; }
			else
				process-prereq-list-remote || { log.error "$(eval_gettext "setup_update_os_failed \$HOSTNAME_FQDN \$HOST_TYPE")" && echo; return 1; }
			fi
		fi
	done < <(read-host-roles)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

get-docker-type() {
	local rc=0
	local -r docker_type=$(get-property-value "$PROPERTY_FILE" "docker.type" "ce")

	case $docker_type in
	ce)
		;;
	ee)
		;;
	*)
		rc=1
		;;
	esac

	echo "$docker_type"

	return $rc
}

get-docker-package() {
	local docker_type="$1"; shift
	local rc=0
	local docker_package=""
	local docker_version="$(get-property-value "${PROPERTY_FILE}" "version.docker")"
	if [[ "$docker_type" == "ce" && "$docker_version" == "" ]]; then
		docker_version=18.03.1.ce
	elif [[ "$docker_type" == "ee" && "$docker_version" == "" ]]; then
		docker_version=17.06.2.ee.6
	fi

	case $docker_type in
	ce)
		docker_package="docker-ce-${docker_version}"
		;;
	ee)
		docker_package="docker-ee-${docker_version}"
		;;
	*)
		rc=1
		;;
	esac

	echo "$docker_package"

	return $rc
}

get-docker-repo-url() {
	local docker_type="$1"; shift
	local rc=0
	local docker_repository=""

	case $docker_type in
	ce)
		docker_repository="https://download.docker.com/linux/centos/docker-ce.repo"
		;;
	ee)
		local -r docker_ee_url=$(get-property-value "$PROPERTY_FILE" "docker.ee.url")
		docker_repository="${docker_ee_url}"
		;;
	*)
		rc=1
		;;
	esac

	echo "$docker_repository"

	return $rc
}

verifyPrereqPackages() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_verify_prereq_packages"

	while read HOST_TYPE; do
		if [[ $HOST_TYPE == docker ]] || [[ $HOST_TYPE == worker* ]]; then	
			local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
			scp "${SCRIPT_DIR}/prepare/verifyPrereqPackages.txt" \
				"${HOSTNAME_FQDN}:."
			ssh -n $HOSTNAME_FQDN "bash verifyPrereqPackages.txt; export RC=$?; rm verifyPrereqPackages.txt; exit $RC"
		fi
	done < <(read-host-roles)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

setupYumRepo() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local -r where="$1"; shift
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.${where}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_setup_yum_repo_$where"

	local -r hostname=$(hostname -f)
	local -r rpm_dir="$SCRIPT_DIR/distrib"

	log.info "$(gettext -s "step_setup_yum_repo_offline")"
	while read HOSTNAME_FQDN; do
		if [[ $where == "local" && $HOSTNAME_FQDN == "$hostname" ]] || \
		   [[ $where == "remote" && $HOSTNAME_FQDN != "$hostname" ]]; then
			if [[ $HOSTNAME_FQDN != "$hostname" ]]; then
				scp -r "$rpm_dir" "$HOSTNAME_FQDN:."
				# shellcheck disable=SC2029
				ssh -n "$HOSTNAME_FQDN" "pushd $(basename "$rpm_dir"); yum -y install --nogpgcheck --disablerepo='*' *.rpm; popd; /bin/rm -fr $(basename "$rpm_dir")" || { log.error "$(eval_gettext "step_setup_yum_repo_setup_failed \$HOSTNAME_FQDN")" && echo; return 1; }
			else
				pushd "$rpm_dir"
				yum -y install --nogpgcheck --disablerepo='*' ./*.rpm || { log.error "$(eval_gettext "step_setup_yum_repo_setup_failed \$HOSTNAME_FQDN")" && echo; return 1; }
				popd
			fi
		fi
	done < <(read-host-fqdn)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

installDocker() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	# Install docker
	log.header "step_install_docker"

	local -r docker_type=$( get-docker-type )
	local -r -i docker_type_rc=$?
	local -r docker_package=$( get-docker-package "$docker_type" )
	local -r -i docker_package_rc=$?

	[[ $docker_type_rc -eq 0 ]] || error_count=$((error_count + 1)) && echo "$docker_type"
	[[ $docker_package_rc -eq 0 ]] || error_count=$((error_count + 1)) && echo "$docker_package"

	if [[ $error_count -gt 0 ]]; then
		return $error_count
	fi

	while read HOST_TYPE; do
		if [[ $HOST_TYPE == docker ]] || \
		   [[ $HOST_TYPE == manager ]] || \
		   [[ $HOST_TYPE == worker* ]]; then
			local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
			log.info "$(eval_gettext "step_install_docker_server \$HOSTNAME_FQDN")"
			if [[ $HOSTNAME_FQDN != $(hostname -f) ]]; then
				# Upload files to host and run
				scp "${SCRIPT_DIR}/prepare/install_docker.txt" "root@$HOSTNAME_FQDN:."
				# shellcheck disable=SC2029
				ssh -n "root@$HOSTNAME_FQDN" "bash install_docker.txt $docker_package && rm install_docker.txt" ||  { log.error "$(eval_gettext "step_install_docker_server_failed \$HOSTNAME_FQDN")" && echo; return 1; }

			else
				bash "${SCRIPT_DIR}/prepare/install_docker.txt" "$docker_package" || { log.error "$(eval_gettext "step_install_docker_server_failed \$HOSTNAME_FQDN")" && echo; return 1; }
			fi
		fi
	done < <(read-host-roles)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

get-docker-registry-url() {
	local docker_registry_url=$(get-property-value "${PROPERTY_FILE}" "external.docker.registry.url")

        if [[ $docker_registry_url == "" ]]; then
                local -a acceptable=(docker manager)
                local docker_host=$(find-best-host acceptable[@])
                if [[ $docker_host != "" ]]; then
                        docker_registry_url="$docker_host:5000"
                fi
        fi

	echo "$docker_registry_url"
}

configureInsecureDockerRegistry() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local -r external_registry_is_secure=$(get-property-value "$PROPERTY_FILE" "external.docker.registry.isSecure")

	if [[ ${external_registry_is_secure,,} == "true" ]] || [[ $external_registry_is_secure == "1" ]]; then
		:
	else
		# Configure insecure docker registry
		log.header "step_configure_docker_reg_security"
		local registry_url=$(get-docker-registry-url)

		local hostname="$(hostname -f)"
		while read HOST_TYPE; do
			if [[ $HOST_TYPE == manager ]] || [[ $HOST_TYPE == worker* ]]; then
				local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
				log.info "$(eval_gettext "step_configure_docker_reg_security_server \$HOSTNAME_FQDN")"
				if [[ $HOSTNAME_FQDN != "$hostname" ]]; then
					# Upload files to host and run
					scp "${SCRIPT_DIR}/prepare/configure_access_insecure_docker_registry.txt" "root@$HOSTNAME_FQDN:."
					# shellcheck disable=SC2029
					ssh -n "root@$HOSTNAME_FQDN" "bash configure_access_insecure_docker_registry.txt $registry_url && rm configure_access_insecure_docker_registry.txt" ||  { log.error "$(eval_gettext "step_configure_docker_reg_security_server_failed \$HOSTNAME_FQDN")" && echo; return 1; }
					echo
				else
					bash "${SCRIPT_DIR}/prepare/configure_access_insecure_docker_registry.txt" "$registry_url" || { log.error "$(eval_gettext "step_configure_docker_reg_security_server_failed \$HOSTNAME_FQDN")" && echo; return 1; }
				fi
			fi
		done < <(read-host-roles)

	fi

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

find-docker-image() {
	local -r image="$1"; shift
	local -r image_dir="$1"; shift
	local -i rc=1

	for tar_file in "${image_dir}"/*.tar; do
		tar xOf "$tar_file" manifest.json | tr "\"" "\n" | grep -q "$image"
		if [[ $? -eq 0 ]]; then
			rc=0
			echo "${tar_file##*/}"
			break;
		fi
	done

	return $rc
}

loadDockerImages() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local docker_registry_url=$(get-docker-registry-url)
	local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")
	local -a images=()
	local counter=1

	local -r image_dir="$SCRIPT_DIR/distrib"
	for image_file in "${image_dir}"/*.tar; do
		log.info "$(eval_gettext "step_load_docker_image \$image_file")"
		local image=$(docker load -i "$image_file" | grep 'Loaded image:' | cut -c15-)
		if [[ "${PIPESTATUS[0]}" != "0" ]]; then
			log.error "$(eval_gettext "step_load_docker_image_failed \$image_file")"
			error_count=$((error_count + 1))
			image=""
		fi
		if [[ $image != "" ]]; then
			images+=($image)
		fi
	done

	if [[ $error_count -gt 0 ]]; then
		return 1
	fi

	for image in "${images[@]}"; do
		local image_id=$(docker inspect --format="{{.Id}}" "$image")
		if [[ $image_id != "" ]]; then
			# If image has a hostname (gcr.io / quay.io / ...) in
			# it, we need to remove it.
			local image_front="${image%%/*}"
			if [[ $image_front != "" ]]; then
				if [[ ${image_front%%*.*} == "" ]]; then
					image="${image#*/}"
				fi
			fi
			docker tag "$image_id" "${docker_registry_url}/${image}"
			if [[ $? -eq 0 ]]; then
				local retry_count=5
				while [[ $retry_count -gt 0 ]]; do
					docker push "${docker_registry_url}/${image}"
					if [[ $? -eq 0 ]]; then
						log.info "$(eval_gettext "step_load_docker_image_push_pass \$docker_registry_url \$image")"
						break;
					else
						retry_count=$((retry_count - 1))
						log.warn "$(eval_ngettext "step_load_docker_image_push_fail \$docker_registry_url \$image \$retry_count" "step_load_docker_image_push_fails \$docker_registry_url \$image \$retry_count" $retry_count)"
					fi
				done
				if [[ $retry_count -eq 0 ]]; then
					log.error "$(eval_gettext "step_load_docker_image_push_fatal \$docker_registry_url \$image")"
					error_count=$((error_count + 1))
				fi
			else
				log.error "$(eval_gettext "step_load_docker_image_tag_failed \$image_id \$docker_registry_url \$image")"
				error_count=$((error_count + 1))
			fi
		else
			log.error "$(eval_gettext "step_load_docker_image_id_failed \$image")"
			error_count=$((error_count + 1))
		fi
	done

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

# There are three scenarios:
# A) Installing new registry for the cluster that the cluster hosts.
# B) Customer has a registry, but we need to push images into it.
# C) Customer has a registry, and they preloaded the required images in it.
installDockerRegistry() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local -r external_registry=$(get-property-value "$PROPERTY_FILE" "external.docker.registry.url")

	local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")
	local -r repo_dir=$(get-property-value "${PROPERTY_FILE}" "docker.registry.repo_dir" "/docker-repo")
	local -r repo_tar=$(get-property-value "${PROPERTY_FILE}" "docker.registry.file")
	local image_dir="$SCRIPT_DIR/distrib"
	local -r registry_image="registry:2"
	local registry_image_tar=""

	if [[ $external_registry == "" ]]; then
		# "A" -  We manage the registry
		local -a acceptable=(docker manager)
		local docker_host=$(find-best-host acceptable[@])

		log.header "step_install_docker_reg \$docker_host"

		registry_image_tar=$(find-docker-image "$registry_image" "$image_dir") || { \
			image=$registry_image
			log.error "$(eval_gettext "find_docker_image_archive_failed \$image \$image_dir")" && echo; \
			return 1; \
		}

		if [[ $local_docker -eq 0 ]]; then
			if [[ $registry_image_tar != "" ]]; then
				scp "${image_dir}/${registry_image_tar}" "root@$docker_host:."
			fi
			scp "${media_dir}/archives/${repo_tar}" "root@$docker_host:."
			scp "${SCRIPT_DIR}/prepare/install_docker_registry.txt" "root@$docker_host:."
			# shellcheck disable=SC2029
			ssh -n "root@$docker_host" \
				"bash install_docker_registry.txt 5000 \"$repo_dir\" \"$repo_tar\" \"$registry_image\" \"./${registry_image_tar}\" && \
				rm -f \"${repo_tar}\" install_docker_registry.txt \"$registry_image\"" || { \
					log.error "$(eval_gettext "step_install_docker_reg_failed \$docker_host")" && echo; \
					return 1; \
				}
		else
			bash "${SCRIPT_DIR}/prepare/install_docker_registry.txt" "${repo_dir}" "${repo_tar}" "$registry_image" "${image_dir}/${registry_image_tar}" || { \
				log.error "$(eval_gettext "step_install_docker_reg_failed \$docker_host")" && echo; \
				return 1; \
			}
		fi
	else
		local -r preloaded=$(get-property-value "$PROPERTY_FILE" "external.docker.registry.isPreloaded")
		if [[ ${preloaded,,} == "true" ]] || [[ $preloaded == "1" ]]; then
			# "C" -- Nothing to do.
			:
		else
			# "B" -- We need to stand up a temp registry, copy images, and tear it down.
			# Always use localhost for docker image to get around
			# configuring insecure registry stuff temporarily.
			bash "${SCRIPT_DIR}/prepare/copy_images_to_other_repository.txt" "${external_registry}" "${repo_dir}-inst-k8s-docker-reg" "${media_dir}/archives/${repo_tar}" "$registry_image" "${image_dir}/${registry_image_tar}" || { \
				log.error "$(eval_gettext "step_install_docker_reg_failed \$docker_host")" && echo; \
				return 1; \
			}
		fi
	fi

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

# work around for https://github.com/kubernetes/kubeadm/issues/1003
fix-k8s-gcr-io-pause-amd64-3-1() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_fix_kubeadm_1003"

	# Give the stuff we've done up to this point some time to settle down
        # before modifying the docker registry...
	sleep 15
	local registry_url=$(get-docker-registry-url)
	local pull="docker pull $registry_url/pause-amd64:3.1"
	local tag="docker tag $registry_url/pause-amd64:3.1 k8s.gcr.io/pause-amd64:3.1"
	echo $pull
	echo $tag
	while read HOST_TYPE; do
		if [[ $HOST_TYPE == manager ]] || \
		   [[ $HOST_TYPE == worker* ]]; then
			local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
			if [[ $HOSTNAME_FQDN != $(hostname -f) ]]; then
				ssh -n "root@$HOSTNAME_FQDN" "$pull && $tag" ||  { log.error "$(eval_gettext "step_fix_kubeadm_1003_failed \$HOSTNAME_FQDN")" && echo; return 1; }
			else
				#$( $pull && $tag ) || { log.error "$(eval_gettext "step_fix_kubeadm_1003_failed \$HOSTNAME_FQDN")" && echo; return 1; }
				eval $pull || { log.error "$(eval_gettext "step_fix_kubeadm_1003_failed \$HOSTNAME_FQDN")" && echo; return 1; }
				eval $tag || { log.error "$(eval_gettext "step_fix_kubeadm_1003_failed \$HOSTNAME_FQDN")" && echo; return 1; }
			fi
		fi
	done < <(read-host-roles)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

installKubernetes() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_install_k8s"
	local -r hostname="$(hostname -f)"
	local kube_version="$(get-property-value "${PROPERTY_FILE}" "version.kubernetes" "1.10.1")"
	local kube_cni_version="$(get-property-value "${PROPERTY_FILE}" "version.kubernetes.cni" "0.6.0")"

	while read HOST_TYPE; do
		if [[ $HOST_TYPE == manager ]] || [[ $HOST_TYPE == worker* ]] ; then
			local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
			log.info "$(eval_gettext "step_install_k8s_server \$HOSTNAME_FQDN")"
			if [[ $HOSTNAME_FQDN != "$hostname" ]]; then
				# Upload files to host and run
				scp "${SCRIPT_DIR}/prepare/install_kubernetes.txt" "root@$HOSTNAME_FQDN:."
				# shellcheck disable=SC2029
				ssh -n "root@$HOSTNAME_FQDN" "bash install_kubernetes.txt $kube_version $kube_cni_version && rm install_kubernetes.txt" ||  { log.error "$(eval_gettext "step_install_k8s_server_fail \$HOSTNAME_FQDN")" && echo; return 1; }

			else
				bash "${SCRIPT_DIR}/prepare/install_kubernetes.txt" "$kube_version" "$kube_cni_version" || { log.error "$(eval_gettext "step_install_k8s_server_fail \$HOSTNAME_FQDN")" && echo; return 1; }
			fi
		fi
	done < <(read-host-roles)

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

installHelm() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	command -v helm > /dev/null

	if [[ $? -ne 0 ]]; then
		log.header "step_install_helm"

		local helm_version="$(get-property-value "${PROPERTY_FILE}" "version.helm" "2.9.0")"
		local helm_bin="helm-v${helm_version}-linux-amd64.tar.gz"
		local -r helm_dest="/usr/local/bin/helm"
		local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")
		local -r rpm_dir="$SCRIPT_DIR/distrib"
		/bin/cp "${rpm_dir}/${helm_bin}" "${media_dir}/archives/${helm_bin}"
		mkdir -p "$media_dir/archives/helm"
		tar xzf "$media_dir/archives/${helm_bin}" -C "$media_dir/archives/helm" || {
			log.error "$(eval_gettext "step_extracting_helm_failed \$helm_bin")"
			echo
			return 1
		}
		/bin/cp "$media_dir/archives/helm/linux-amd64/helm" "${helm_dest}" || {
			log.error "$(eval_gettext "step_install_helm_failed")"
			echo
			return 1
		}
		/bin/rm -fr "$media_dir/archives/helm"
	fi

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

initializeHelm() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local -r helm_cmd=$(command -v helm)
	local helm_version="$(get-property-value "${PROPERTY_FILE}" "version.helm" "2.9.0")"

	local have_tiller=$(kubectl -n kube-system get pod -l app=helm,name=tiller | wc -l)
	[[ $have_tiller -eq 0 ]] || return 0

	local tiller_image_url="gcr.io/kubernetes-helm/tiller:v${helm_version}"
	local repository=$(get-docker-registry-url)
	tiller_image_url="$repository/kubernetes-helm/tiller:v${helm_version}"

	kubectl create serviceaccount --namespace kube-system helm || {
		log.error "$(eval_gettext "step_configure_helm_failed")"
		echo
		return 1
	}

	kubectl create clusterrolebinding helm-cluster-rule \
		--clusterrole=cluster-admin \
		--serviceaccount=kube-system:helm || {
		log.error "$(eval_gettext "step_configure_helm_failed")"
		echo
		return 1
	}

	"${helm_cmd}" init --wait --service-account helm --skip-refresh --tiller-image "$tiller_image_url" || {
		log.error "$(eval_gettext "step_init_helm_failed")"
		echo
		return 1
	}

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

installNFSServer() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local config_nfs=$(get-property-value "${PROPERTY_FILE}" "configure.nfsserver")
	local configure_nfs
	local external_nfs_server=$(get-property-value "${PROPERTY_FILE}" "external.nfsserver")

	if [[ ${config_nfs,,} == "false" ]] || [[ $config_nfs == "0" ]]; then
		configure_nfs=0
	else
		configure_nfs=1
	fi

	if [[ $external_nfs_server == "" && $configure_nfs == 1 ]]; then
		local -r master_mount=$(get-property-value "${PROPERTY_FILE}" "nfsserver.master.mount")
		log.header "step_create_nfs"

		local hostname=$(hostname -f)
		local -a acceptable=(nfs manager)
		local nfs_server=$(find-best-host acceptable[@]) local_nfs=0

		if [[ $nfs_server == "$hostname" ]]; then
			local_nfs=1
		fi

		if [[ $local_nfs -eq 0 ]]; then
			scp "${SCRIPT_DIR}/prepare/install_configure_nfs_server.txt" "root@$nfs_server:."
			# shellcheck disable=SC2029
			ssh -n "root@$nfs_server" "bash install_configure_nfs_server.txt $master_mount && \
				rm install_configure_nfs_server.txt" || { \
					log.error "$(eval_gettext "step_create_nfs_failed \$nfs_server")" && echo; \
					return 1; \
				}
		else
			bash "${SCRIPT_DIR}/prepare/install_configure_nfs_server.txt" "$master_mount" || { \
				log.error "$(eval_gettext "step_create_nfs_failed \$nfs_server")" && echo; \
				return 1; \
			}
		fi
	fi

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

create-mount-point() {
	# Used in logging...
	# shellcheck disable=SC2034
	local -r index="${1}"; shift;
	local -r mount_parent="${1}"; shift;
	local path owner="root:root" permissions="777"
	while [[ $# -gt 0 ]]; do
		local key="$1"; shift
		case $key in
		--path)
			path="$1"; shift
			;;
		--owner)
			owner="$1"; shift
			;;
		--permissions)
			permissions="$1"; shift
			;;
		*)
			log.error "$(eval_gettext "error_unrecognized_parameter \$key")"
                	return 1
			;;
		esac
	done

	if [[ "$path" != "" ]]; then
		if ! mkdir -p "${mount_parent}${path}"; then
			log.error "$(eval_gettext "step_ensure_nfs_mounts_create_mount_failed \$mount_parent \$path")"
			return 1
		fi
		if ! chown "$owner" "${mount_parent}${path}"; then
			log.error "$(eval_gettext "step_ensure_nfs_mounts_chown_mount_failed \$mount_parent \$path \$owner")"
			return 1
		fi
		if ! chmod "$permissions" "${mount_parent}${path}"; then
			log.error "$(eval_gettext "step_ensure_nfs_mounts_chmod_mount_failed \$mount_parent \$path \$permissions")"
			return 1
		fi
	else
		log.error "$(eval_gettext -s "create_mount_no_path \$index")"
		return 1
	fi

	echo "$path"

	return 0
}

ensureNFSMountPoints() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_ensure_nfs_mounts"

	local local_nfs=0 nfs_server=$(get-property-value "${PROPERTY_FILE}" "external.nfsserver")

	if [[ $nfs_server == "" ]]; then
		# The cluster controls the NFS server.

		local -a acceptable=(nfs manager)
		nfs_server=$(find-best-host acceptable[@])

		if [[ $nfs_server == "$hostname" ]]; then
			local_nfs=1
		fi
	fi
	log.info "$(eval_gettext "step_ensure_nfs_mounts \$nfs_server")"

	local test_mnt="${SCRIPT_DIR}/prepare/_generatedScripts/mnt"
	mkdir -p "$test_mnt"
	# Mount the root of the filesystem and create the directories needed
	mount -t nfs -o rw,vers=4.1,proto=tcp "${nfs_server}:/" "$test_mnt" || { \
		log.error "$(eval_gettext "step_ensure_nfs_mounts_mount_failed \$nfs_server \$test_mnt")" && echo; \
		return 1; \
	}

	# Create the mount points
	local counter=1 path
	local -a mount_points=()
	local mount_point=$(get-property-value "${PROPERTY_FILE}" "mount_point.$counter")
	while [[ $mount_point != "" ]]; do
		# shellcheck disable=SC2086
		path=$(create-mount-point "$counter" "$test_mnt" $mount_point)
		if [[ "$?" == 0 ]]; then
			mount_points+=($path)
		else
			error_count=$((error_count + 1))
		fi
		counter=$((counter + 1))
		mount_point=$(get-property-value "${PROPERTY_FILE}" "mount_point.$counter")
	done

	umount "$test_mnt" || {
		log.error "$(eval_gettext "step_ensure_nfs_mounts_umount_failed \$nfs_server \$test_mnt")"
		echo
		 return 1;
	}

	log.info "$(gettext -s "step_ensure_nfs_mounts_can_mount")"
	for mount_point in "${mount_points[@]}"; do
		mount -t nfs -o rw,vers=4.1,proto=tcp "${nfs_server}:${mount_point}" \
			"$test_mnt" && \
			ls "$test_mnt" && \
			sleep 2 && \
			umount "$test_mnt"
		if [[ $? != 0 ]]; then
			log.error "$(eval_gettext "step_ensure_nfs_mounts_can_mount_failed \$nfs_server \$mount_point")"
			error_count=$((error_count + 1))
		fi
	done
	rmdir "$test_mnt"

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

apply-node-taint() {
	local error_count=0 node="" taint_key="" value="" effect=""

	while [[ $# -gt 0 ]]; do
		local key="$1"; shift
		case $key in
		-node)
			node="$1"; shift
			;;
		-key)
			taint_key="$1"; shift
			;;
		-value)
			value="$1"; shift
			;;
		-effect)
			effect="$1"; shift
			;;
		*)
			log.error "$(eval_gettext "error_unrecognized_parameter \$key")"
			return 1
			;;
		esac
	done

	if [[ $node == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_node")"
		error_count=$((error_count + 1))
	fi

	if [[ $taint_key == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_key")"
		error_count=$((error_count + 1))
	fi

	if [[ $value == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_value")"
		error_count=$((error_count + 1))
	fi

	if [[ $effect == "" ]]; then
		log.error "$(gettext -s "apply_taint_no_effect")"
		error_count=$((error_count + 1))
	fi

	case $effect in
	NoSchedule)
		;;
	PreferNoSchedule)
		;;
	NoExecute)
		;;
	*)
		log.error "$(eval_gettext "apply_taint_unrecognized_effect \$effect")"
		error_count=$((error_count + 1))
		;;
	esac

	[[ $error_count -gt 0 ]] && return $error_count

	local -a acceptable=($node)
	local node_fqdn=$(find-best-host acceptable[@])

	if [[ $node_fqdn == "" ]]; then
		log.error "$(gettext -s "apply_taint_node_not_found \$node")"
		error_count=$((error_count + 1))
	fi

	[[ $error_count -gt 0 ]] && return $error_count

	kubectl taint nodes "$node_fqdn" "$taint_key=$value:$effect"

	return $?
}

apply-node-taints() {
	local error_count=0 taint_index=1

	local taint=$(get-property-value "${PROPERTY_FILE}" "taint.$taint_index")
	while [[ $taint != "" ]]; do
		# shellcheck disable=SC2086
		apply-node-taint $taint
		taint_index=$((taint_index + 1))
		taint=$(get-property-value "${PROPERTY_FILE}" "taint.$taint_index")
	done

	return $error_count
}

apply-node-label() {
	local error_count=0 node="" label_key="" value="" effect=""

	while [[ $# -gt 0 ]]; do
		local key="$1"; shift
		case $key in
		-node)
			node="$1"; shift
			;;
		-key)
			label_key="$1"; shift
			;;
		-value)
			value="$1"; shift
			;;
		*)
			log.error "$(eval_gettext "error_unrecognized_parameter \$key")"
			return 1
			;;
		esac
	done

	if [[ $node == "" ]]; then
		log.error "$(gettext -s "apply_label_no_node")"
		error_count=$((error_count + 1))
	fi

	if [[ $label_key == "" ]]; then
		log.error "$(gettext -s "apply_label_no_key")"
		error_count=$((error_count + 1))
	fi

	if [[ $value == "" ]]; then
		log.error "$(gettext -s "apply_label_no_value")"
		error_count=$((error_count + 1))
	fi

	[[ $error_count -gt 0 ]] && return $error_count

	local -a acceptable=($node)
	local node_fqdn=$(find-best-host acceptable[@])

	if [[ $node_fqdn == "" ]]; then
		log.error "$(gettext -s "apply_label_node_not_found \$node")"
		error_count=$((error_count + 1))
	fi

	[[ $error_count -gt 0 ]] && return $error_count

	kubectl label nodes "$node_fqdn" "$label_key=$value"

	return $?
}

apply-node-labels() {
	local error_count=0 label_index=1

	local label=$(get-property-value "${PROPERTY_FILE}" "label.$label_index")
	while [[ $label != "" ]]; do
		# shellcheck disable=SC2086
		apply-node-label $label
		label_index=$((label_index + 1))
		label=$(get-property-value "${PROPERTY_FILE}" "label.$label_index")
	done

}

createKubeCluster() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	log.header "step_create_kube_cluster"
	local manager="" manager_fqdn="" delay_block_size=15 max_delay_block=20
	local hostname="$(hostname -f)"
	local repository="$(get-docker-registry-url)"
	local calicoctl_loc=""

	# If a cluster already exists, we'll use it, don't rebuild it.
	if kubectl get nodes; then
		return 0
	fi

	manager="$(get-property-value "${HOSTNAME_FILE}" "manager.ip")"
	# Used in logging...
	# shellcheck disable=SC2034
	manager_fqdn="$(get-property-value "${HOSTNAME_FILE}" "manager.fqdn")"

	local -r pod_cidr=$(get-property-value "${PROPERTY_FILE}" "k8s.pod_cidr")
	local -r svc_cidr=$(get-property-value "${PROPERTY_FILE}" "k8s.svc_cidr")
	local -r dns_ip=$(get-property-value "${PROPERTY_FILE}" "k8s.dns_ip")
	local -r calico_etcd_ip=$(get-property-value "${PROPERTY_FILE}" "k8s.calico_etcd_ip" "$manager")

	local rpm_dir="$SCRIPT_DIR/distrib"
	calicoctl_loc="${rpm_dir}/calicoctl"

	local -i mtu=$(cat "/sys/class/net/$(ip address | grep "$manager" | awk '{print $NF}')/mtu")
	if [[ $mtu -gt 110 ]]; then
		mtu=$((mtu - 110))
	else
		mtu=1390
	fi

	local kube_version="$(get-property-value "${PROPERTY_FILE}" "version.kubernetes" "1.10.1")"
	bash "${SCRIPT_DIR}/prepare/initialize_kubernetes_master.txt" "$manager" "$delay_block_size" "$max_delay_block" "$repository" "$calicoctl_loc" "$pod_cidr" "$svc_cidr" "$dns_ip" "$calico_etcd_ip" "$mtu" "$kube_version" || { log.error "$(eval_gettext "step_create_kube_cluster_fail \$HOSTNAME_FQDN")" && echo; return 1; }
	wait-for-kube-manager-ready

	local -r token=$(kubeadm token list | grep bootstrap | awk '{print $1}')
	local -r discovery_hash=$(openssl x509 -pubkey -in /etc/kubernetes/pki/ca.crt | openssl rsa -pubin -outform der 2>/dev/null | openssl dgst -sha256 -hex | sed 's/^.* //')
	local worker_count=0

	while read HOST_TYPE; do
		if [[ $HOST_TYPE == worker* ]] ; then
			local HOSTNAME_FQDN="$(get-property-value "${HOSTNAME_FILE}" "${HOST_TYPE}.fqdn")"
			worker_count=$((worker_count + 1))
			log.info "$(eval_gettext "step_create_kube_cluster_join_worker \$HOSTNAME_FQDN")"
			# Upload files to host and run
			scp "${SCRIPT_DIR}/prepare/join_worker.txt" "root@$HOSTNAME_FQDN:."
			# shellcheck disable=SC2029
			ssh -n "root@$HOSTNAME_FQDN" "bash join_worker.txt \"$manager\" \"$token\" \"$discovery_hash\" \"$repository\" \"$dns_ip\" \"$kube_version\" && rm join_worker.txt" ||  { log.error "$(eval_gettext "step_create_kube_cluster_join_worker_fail \$HOSTNAME_FQDN")" && echo; return 1; }

		fi
	done < <(read-host-roles)

	if [[ $worker_count == "0" ]]; then
		log.info "$(gettext -s "step_create_kube_cluster_no_worker")"
		case "$kube_version" in
		1.7.*)
			kubectl taint nodes --all node-role.kubernetes.io/master- || { log.error "$(eval_gettext "step_create_kube_cluster_no_worker_remove_taint_failed \$manager_fqdn")" && echo; error_count=$((error_count + 1)); return 1; }
			;;
		1.8.*)
			kubectl taint nodes --all node-role.kubernetes.io/master- || { log.error "$(eval_gettext "step_create_kube_cluster_no_worker_remove_taint_failed \$manager_fqdn")" && echo; error_count=$((error_count + 1)); return 1; }
			;;
		1.9.*)
			kubectl taint nodes --all node-role.kubernetes.io/master- || { log.error "$(eval_gettext "step_create_kube_cluster_no_worker_remove_taint_failed \$manager_fqdn")" && echo; error_count=$((error_count + 1)); return 1; }
			;;
		1.10.*)
			kubectl taint nodes --all node-role.kubernetes.io/master- || { log.error "$(eval_gettext "step_create_kube_cluster_no_worker_remove_taint_failed \$manager_fqdn")" && echo; error_count=$((error_count + 1)); return 1; }
			;;
		*)
			log.error "$(eval_gettext "step_create_kube_cluster_no_worker_remove_taint_failed \$manager_fqdn")"
			return 1
		esac
	else
		# wait for all of the worker nodes to be ready.
		local -r ready_node_count=$((worker_count + 1))
		wait-for-all-nodes-ready "$ready_node_count" || error_count=$((error_count + 1))
		apply-node-taints || error_count=$((error_count + 1))
	fi
	echo

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

verifyClusterAdministrator() {
        INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
        local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
        is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

        log.header "step_verify_cluster_admin"
        local kube_user="$(get-property-value "${PROPERTY_FILE}" "k8s.user")"
        local kube_passwd="$(get-property-value "${PROPERTY_FILE}" "k8s.password")"
        local kube_group="$(get-property-value "${PROPERTY_FILE}" "k8s.group")"
        if [[ "$kube_user" != "" ]]; then
                id "$kube_user" > /dev/null
                if [[ $? -ne 0 ]]; then
                	log.info "$(eval_gettext "step_verify_cluster_admin_creating_user \$kube_user")"
                        adduser "$kube_user" || { log.error "$(eval_gettext "step_verify_cluster_admin_adduser \$kube_user")" && echo; return 1; }
			if [[ "$kube_passwd" != "" ]]; then
                		log.info "$(eval_gettext "step_verify_cluster_admin_setting_password \$kube_user")"
				echo "$kube_passwd" | passwd -f --stdin "$kube_user" || { log.error "$(eval_gettext "step_verify_cluster_admin_passwd \$kube_user")" && echo; return 1; }
			fi
                fi
                if [[ "$kube_group" != "" ]]; then
                        if [ ! "$(getent group "$kube_group")" ]; then
                		log.info "$(eval_gettext "step_verify_cluster_admin_creating_group \$kube_group")"
                                groupadd "$kube_group" || { log.error "$(eval_gettext "step_verify_cluster_admin_groupadd \$kube_group")" && echo; return 1; }
                        fi
                        if ! id -nG | grep -qw "$kube_group" ; then
                		log.info "$(eval_gettext "step_verify_cluster_admin_add_user_group \$kube_user \$kube_group")"
                                usermod -aG "$kube_group" "$kube_user" || { log.error "$(eval_gettext "step_verify_cluster_admin_usermod \$kube_user \$kube_group")" && echo; return 1; }
                        fi
                fi
		local admin_home=$(getent passwd "$kube_user" | cut -d: -f6)
		local admin_group="$kube_group"
		if [[ "$admin_group" == "" ]]; then
			admin_group="$kube_user"
		fi
                log.info "$(eval_gettext "step_verify_cluster_admin_copy_cluster_config")"
		mkdir "$admin_home/.kube" && \
		cp /etc/kubernetes/admin.conf "$admin_home/.kube/config" && \
		chown -R "$kube_user:$admin_group" "$admin_home/.kube" && \
		chmod 700 "$admin_home/.kube" && \
		chmod 600 "$admin_home/.kube/config"
		if [[ $? -ne 0 ]]; then
			log.error "$(eval_gettext "step_verify_cluster_admin_copy_cluster_config_failed")"
			echo
			return 1
		fi
        else
                log.info "$(eval_gettext "step_verify_cluster_admin_not_configured")"
        fi

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

get-chart-pv-arg() {
	local pvArg
	local config_nfs=$(get-property-value "${PROPERTY_FILE}" "configure.nfsserver" "true")
	if [[ ${config_nfs,,} == "true" ]] || [[ $config_nfs == "1" ]]; then
		local nfs_server=$(get-property-value "${PROPERTY_FILE}" "external.nfsserver")
		if [[ $nfs_server == "" ]]; then
			local -a acceptable=(nfs manager)
			nfs_server=$(find-best-host acceptable[@]) local_nfs=0
		fi
		pvArg="global.nfsServer=$nfs_server"
	else
		pvArg="global.createPv=false"
	fi
	echo $pvArg
}

get-chart-docker-arg() {
	echo "global.dockerRepository=$(get-docker-registry-url)"
}

enable-port-forwarding() {
	local install_result="$1"; shift
	local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")
	local proxy_file=$(find "$media_dir/archives" -name "$proxy_chart")

set -x
	if [[ $proxy_file != "" ]]; then
		local manager_ip="$(get-property-value "${HOSTNAME_FILE}" "manager.ip")"
		local external_manager_ip="$(get-property-value "${HOSTNAME_FILE}" "manager.external_ip")"
 		local managerArg

		if [[ $external_manager_ip != "" ]]; then
			managerArg="managerIPAddresses={$manager_ip,$external_manager_ip}"
		else
			managerArg="managerIPAddresses={$manager_ip}"
		fi
		tmp="${install_result##*== Common port forwarding ==}"
		local port_forwards="${tmp%%== *}"
		port_forwards="${port_forwards%%[[:space:]]}"
		port_forwards="${port_forwards##[[:space:]]}"

		local chart_rel_name="$(get-property-value "${PROPERTY_FILE}" "chart.release.name")"
		local release_name="--name ${chart_rel_name}-proxy"

		if [[ "$port_forwards" != "$install_result" ]]; then

			local i=0 tmpdir=
			trap '[[ $tempdir ]] && rm -fr "$tempdir"' EXIT

			while ((++i <= 10)); do
				tempdir=${TMPDIR:-/tmp}/$RANDOM-$$
				mkdir -m 700 "$tempdir" 2>/dev/null && break
			done
			if ((i > 10)); then
				exit 1
			fi
			cat <<EOF > "$tempdir/values.yaml"
$port_forwards
EOF
			# shellcheck disable=SC2086
			helm install --set "$managerArg" \
				--set "$(get-chart-docker-arg)" \
				$release_name \
				-f "$tempdir/values.yaml" "$proxy_file" || {
				log.error "$(eval_gettext "step_install_chart_install_proxy_failed")"
				log.error "$port_forwards"
				return 1
			}
			echo "helm install return code = $?"
set +x
		fi
	fi
}

initialize-pvs() {
	local install_result="$1"; shift
	local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")
	pushd "$SCRIPT_DIR"

	sleep 15
	tmp="${install_result##*== PV initialization ==}"
	local init_pvs="${tmp%%== *}"
	init_pvs="${init_pvs%%[[:space:]]}"
	init_pvs="${init_pvs##[[:space:]]}"
	init_pvs="${init_pvs//__DATA_FILE_ARCHIVE_DIRECTORY__/$media_dir/archives}"
	if [[ $init_pvs != "" && $init_pvs != "$install_result" ]]; then
		while IF='' read -r line || [[ -n "$line" ]]; do
			line="${line%%[[:space:]]}"
			line="${line##[[:space:]]}"
			if [[ $line == ./initialize-pv* ]]; then
				echo "$line"
set -x
				eval "$line"
set +x
				sleep 2
			fi
		done <<<"$init_pvs"
	fi

	popd
}

installChart() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local tmp
	local chart="$(get-property-value "${PROPERTY_FILE}" "chart")"
	local charts_archive="$(get-property-value "${PROPERTY_FILE}" "charts.archive")"
	local chart_args="$(get-property-value "${PROPERTY_FILE}" "chart.args")"
	local chart_rel_name="$(get-property-value "${PROPERTY_FILE}" "chart.release.name")"
	local -r media_dir=$(get-property-value "${PROPERTY_FILE}" "media.dir")

	if [[ $chart != "" ]]; then
		log.header "step_install_chart"
		local -a acceptable=(manager)
		local manager=$(find-best-host acceptable[@])
		local -a chartArgs

		if [[ $chart_args != "" ]]; then
			chartArgs=($chart_args)
		fi

		if [[ $charts_archive != "" ]]; then
			(cd "$media_dir/archives" && tar xzf "$charts_archive")
		fi

		local chart_file=$(find "$media_dir/archives" -name "$chart")
		local install_result
		if [[ $chart_file != "" ]]; then
			local -r reg_uname=$(get-property-value "${PROPERTY_FILE}" "external.docker.registry.pullsecret.username")
			local -r reg_passwd=$(get-property-value "${PROPERTY_FILE}" "external.docker.registry.pullsecret.password")
			local -r reg_email=$(get-property-value "${PROPERTY_FILE}" "external.docker.registry.pullsecret.email")

			if [[ $reg_uname != "" && $reg_passwd != "" && $reg_email != "" ]]; then
				chartArgs+=(--set "global.externalRegistry.username=$reg_uname")
				chartArgs+=(--set "global.externalRegistry.password=$reg_passwd")
				chartArgs+=(--set "global.externalRegistry.email=$reg_email")
				chartArgs+=(--set "global.externalRegistry.auth=$(echo -n "$reg_uname:$(echo "$reg_passwd" | base64 -d)" | base64)")
			fi
			local repository=$(get-docker-registry-url)
			chartArgs+=(--set "global.initPvImage=$repository/centos:7")

set -x
			install_result=$(helm upgrade --install "$chart_rel_name" \
				"${chartArgs[@]}" \
				--set "global.managerFQDN=$manager" \
				--set "$(get-chart-pv-arg)" \
				--set "$(get-chart-docker-arg)" \
				"$chart_file")
			if [[ $? != 0 ]]; then
				log.error "$(eval_gettext "step_install_chart_chart_install_failed \$chart")"
				log.error "$install_result"
				return 1
			fi
set +x
			enable-port-forwarding "$install_result" || {
				return 1;
			}
			initialize-pvs "$install_result" || {
				return 1;
			}
		else
			log.error "$(eval_gettext "step_install_chart_chart_not_found \$chart")"
		fi

		[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	fi

	return $error_count
}

wait-for-matching-output() {
	local -r max_time="$1"; shift
	local -r message="$1"; shift
	local -r cmd="$1"; shift
	local -r expected="$1"; shift
	local delay_block_size=15 delay_block=0 last_delay_block=0
	local start_time=$(date +%s) end_time
	local elapsed_time

	echo -n "$(eval_gettext "wait_for \$max_time \$message")"

	actual=$(eval "$cmd") || { echo; log.error "$(eval_gettext "wait_for_cmd_error \$cmd")"; return 1; }
	while [[ "$actual" != "$expected" ]]; do
		sleep 1
		end_time=$(date +%s)
		delay_block=$(( (end_time - start_time) / delay_block_size ))

		if [[ $delay_block != $last_delay_block ]]; then
			echo -n "$(gettext "wait_for_progress")"
			last_delay_block=$delay_block
		fi

		if [[ $((end_time-start_time)) -gt $max_time ]]; then
			elapsed_time=$(convert-seconds $((end_time-start_time)) )
			eval_gettext "wait_for_timeout \$elapsed_time"
			return 1
		fi

		actual=$(eval "$cmd") || { echo; log.error "$(eval_gettext "wait_for_cmd_error \$cmd")"; return 1; }
	done
	end_time=$(date +%s)

	# shellcheck disable=SC2034
	elapsed_time=$(convert-seconds $((end_time-start_time)) )
	eval_gettext "wait_for_done \$elapsed_time"
	echo

	return 0
}

wait-for-all-nodes-ready() {
	local -r expected_nodes="$1"; shift
	local -r message="$(eval_gettext "wait_for_cluster_nodes_ready")"

	wait-for-matching-output 900 "$message" \
		"kubectl get nodes | grep -v NotReady | grep Ready | wc -l" \
		"$expected_nodes"

	return $?
}

wait-for-kube-manager-ready() {
	local rc=0
	local message="$(eval_gettext "wait_for_cluster_manager_ready")"

	log.info "$(eval_gettext "wait_for_cluster_manager_kubectl")"
	# Give the manager a bit of time to get started so kubectl works.
	kubectl get nodes > /dev/null
	while [[ $? -ne 0 ]]; do
		sleep 1
		kubectl get nodes > /dev/null
	done
	log.info "$(eval_gettext "wait_for_cluster_manager_kubectl_ok")"
	sleep 1

	wait-for-matching-output 900 "$message" \
		"kubectl get nodes | grep -v NotReady | grep Ready | wc -l" \
		1
	rc=$?

	if [[ $rc -eq 0 ]]; then
		wait-for-pod-ready "-l k8s-app=kube-dns" "kube-system"
		rc=$?
	fi

	return $rc
}

wait-for-pvc-bound() {
	local -r pvc_name="$1"; shift
	local -r message="$(eval_gettext "wait_for_pvc_bound \$pvc_name")"
	local -r target="$(gettext -s "wait_for_pvc_bound_success_string")"

	wait-for-matching-output 300 "$message" \
		"kubectl get pvc $pvc_name -o yaml | grep phase: | awk '{print \$2}'" \
		"$target"

	return $?
}

wait-for-pod-ready() {
	local -r pod="$1"; shift
	local -r namespace="${1:-default}"; shift
	local -r message="$(eval_gettext "wait_for_pod_ready \$pod")"
	local -r target="$(gettext -s "wait_for_pod_ready_success_string")"
	local pod_name=$(kubectl get pod -n "$namespace" "$pod" --output=jsonpath='{$.items[*].metadata.name}')

	if [[ $pod_name == "" ]]; then
		pod_name="$pod"
	fi

	wait-for-matching-output 900 "$message" \
		"kubectl get pod -n $namespace $pod_name --output=jsonpath='{$.status.containerStatuses[*].ready}' | tr ' ' '\n' | sort -u" \
		"$target"

}

initializeKubernetesScripts() {
	INSTALL_STEP="$(gettext -s "${FUNCNAME[0]}")"
	local error_count=0 lockfile="${SCRIPT_DIR}/.${INSTALL_STEP}.lock"
	is_install_step_done "$INSTALL_STEP" "$lockfile" || return 0

	local nfs_server=$(get-property-value "${PROPERTY_FILE}" "external.nfsserver")
	if [[ $nfs_server == "" ]]; then
		# shellcheck disable=SC2034
		local -a acceptable=(nfs manager)
		nfs_server=$(find-best-host acceptable[@])
	fi

	local name="kubernetes-scripts"
	local mount_point="/$name"
	local -r pvc_name=$(create-pv-pvc "$nfs_server" $mount_point \
				--pv "${name}-pv" --sc nfs-storage-class \
				--pvc "${name}-pvc" --app_id "$name" --cap 5Gi) \
	|| { log.error "$(gettext -s "step_init_kube_scripts_create_pv_pvc_failed")" && echo; return 1; }

	wait-for-pvc-bound "$pvc_name" \
	|| { log.error "$(eval_gettext "step_init_kube_scripts_pvc_bind_timeout \$pvc_name")" && echo; return 1; }

	local key="kubernetes-scripts"
	deploy-staging-container "$pvc_name" "$mount_point" || \
	{ log.error "$(eval_gettext "initialize-pvc_staging_deployment_failed \$pvc_name \$mount_point \$key")" && echo; return 1; }

	pod=$(kubectl get pods -l "staging=${pvc_name}" -o jsonpath='{$.items[*].metadata.name}')
	wait-for-pod-ready "$pod" || { log.error "$(eval_gettext "initialize-pvc_pod_not_ready \$pod")" && echo; return 1; }

	log.info "$(gettext -s "initialize-pvc_delete_old_files")"
	kubectl exec -i "$pod" -- sh -c "rm -fr ${mount_point}/*" \
	|| { log.error "$(eval_gettext "initialize-pvc_delete_old_files_failed \$pod \$mount_point")" && echo; return 1; }

	for extra_file in kubernetes-scripts/*; do
		if [[ $extra_file != "" ]]; then
			local extra_dest="$mount_point/${extra_file##*/}"
			log.info "$(eval_gettext "initialize-pvc_cp_extras \$pvc_staging \$extra_file \$pod \$extra_dest")"
			kubectl cp "${SCRIPT_DIR}/${pvc_staging}/${extra_file}" "${pod}:${extra_dest}" \
			|| { log.error "$(eval_gettext "initialize-pvc_cp_extras_failed \$pvc_staging \$extra_file \$pod \$extra_dest")" && echo; return 1; }
			extra_counter=$((extra_counter + 1))
		fi
	done

	perms="755"
	log.info "$(eval_gettext "step_init_kube_scripts_file_perms \$perms")"
	kubectl exec "$pod" -- chmod -R $perms "$mount_point" \
	|| { log.error "$(eval_gettext "step_init_kube_scripts_file_perms_failed \$perms \$mount_point")" && echo; return 1; }

	[[ $error_count -eq 0 ]] && install_step_done "$lockfile"
	return $error_count
}

translations-available() {
	# Make sure translations are up to date.
	# Keep updating key to the last message id added to the catalog.
	local -r key="top_err_fix_kubeadm_1003"
	local -r msg=$(gettext -s "$key")
	if [[ $msg == "$key" ]]; then
		log.error " $(gettext -s "need to run translate.sh")"
		exit 1
	fi
}

# End of function definitions
#################################################################
#								#
#	Main							#
#								#
#################################################################

declare -r -i install_start_time=$(date +%s)

INSTALL_STEP="$(gettext -s "install.sh")"

lockfile="${SCRIPT_DIR}/.install.sh.complete.lock"
is_install_step_done "$INSTALL_STEP" "$lockfile" || exit 0

#translations-available
acceptLicenseAgreement || { log.error "$(gettext -s "top_err_acceptLicenseAgreement")" && echo; exit 1; }
acceptLicenseInformation || { log.error "$(gettext -s "top_err_acceptLicenseInformation")" && echo; exit 1; }

precheckFiles || { log.error "$(gettext -s "top_err_precheckFiles")" && echo; exit 1; }

config_nodes=$(get-property-value "${PROPERTY_FILE}" "configure.nodes" 0)
if [[ ${config_nodes,,} == "true" ]] || [[ $config_nodes == "1" ]]; then
	configure_nodes=1
fi

extractArchives || { log.error "$(gettext -s "top_err_extractArchives")" && echo; exit 1; }
validateHostTypes || { log.error "$(eval_gettext "top_err_validateHostTypes \$HOSTNAME_FILE")" && echo; exit 1; }
pingTest ||  { log.error "$(eval_gettext "top_err_pingTest \$HOSTNAME_FILE")" && echo; exit 1; }
if [[ "$configure_nodes" == "1" ]]; then
	setupYumRepo "local" || { log.error "$(gettext -s "top_err_setupYumRepo")" && echo; exit 1; }
	installExpect ||  { log.error "$(gettext -s "top_err_installExpect")" && echo; exit 1; }
	sshTest || {
        	setupSSHTrust ||  { log.error "$(gettext -s "top_err_setupSSHTrust")" && echo; exit 1; }
        	sshTest || { log.error "$(gettext -s "top_err_sshTest")" && echo; rm -f "${SCRIPT_DIR}/.setupSSHTrust.lock"; exit 1; }
	}
	verifyPrereqPackages || { log.error "$(gettext -s "top_err_verifyPrereqPackages")" && echo; exit 1; }
	setupYumRepo "remote" || { log.error "$(gettext -s "top_err_setupYumRepo")" && echo; exit 1; }
	checkUMASK || { log.error "$(gettext -s "top_err_checkUMASK")" && echo; exit 1; }
	updateOSconfig || { log.error "$(gettext -s "top_err_updateOSconfig")" && echo; exit 1; }
	installDocker || { log.error "$(gettext -s "top_err_installDocker")" && echo; exit 1; }
	installNFSServer || { log.error "$(gettext -s "top_err_installNFSServer")" && echo; exit 1; }
	configureInsecureDockerRegistry || { log.error "$(gettext -s "top_err_configureInsecureDockerRegistry")" && echo; exit 1; }
fi
installDockerRegistry || { log.error "$(gettext -s "top_err_installDockerRegistry")" && echo; exit 1; }
ensureNFSMountPoints || { log.error "$(gettext -s "top_err_ensureNFSMountPoints")" && echo; exit 1; }
loadDockerImages || { log.error "$(gettext -s "top_err_loadDockerImages")" && echo; exit 1; }
if [[ "$configure_nodes" == "1" ]]; then
	fix-k8s-gcr-io-pause-amd64-3-1 || { log.error "$(gettext -s "top_err_fix_kubeadm_1003")" && echo; exit 1; }
	installKubernetes || { log.error "$(gettext -s "top_err_installKubernetes")" && echo; exit 1; }
	installHelm || { log.error "$(gettext -s "top_err_installHelm")" && echo; exit 1; }
	createKubeCluster || { log.error "$(gettext -s "top_err_createKubeCluster")" && echo; exit 1; }
	initializeHelm || { log.error "$(gettext -s "top_err_initializeHelm")" && echo; exit 1; }
	verifyClusterAdministrator || { log.error "$(gettext -s "top_err_createAdditionalUser")" && echo; exit 1; }
fi
if [ -d /etc/cron.daily ]; then
       /bin/cp ${SCRIPT_DIR}/delete_old_jobs /etc/cron.daily/
       chmod 755 /etc/cron.daily/delete_old_jobs
fi

installChart || { log.error "$(gettext -s "top_err_installChart")" && echo; exit 1; }

install_step_done "$lockfile"

declare -r -i install_end_time=$(date +%s)

# shellcheck disable=SC2034
duration=$(convert-seconds $(( install_end_time - install_start_time )) )

INSTALL_STEP="$(gettext -s "install.sh")"
log.info "$(eval_gettext "install.sh.complete \$duration")"

# Remove lock on script.
_remove_lock

exit 0

# Executing this and then executing install.sh again will install the helm chart again.
# Either:
# 1) /bin/rm -fr /fci-exports/*/* /fci-exports/*/.??*
# -- or --
# 2) ctrl-C out of install.sh when it gets to initialing the PVs.

/bin/rm -fr .createKubeCluster.lock .initializeHelm.lock .installChart.lock .install.sh.complete.lock ~/.kube ~/.helm
